import random
import array
import string
import math
import os
import sys
import textwrap
import time
from threading import Thread

from java.lang import System
from java.lang import UnsupportedOperationException
from java.awt import Toolkit
from java.awt import Desktop
from java.net import URI

from org.python.core.util import FileUtil
import java.lang.ClassLoader
import java.lang.Runtime
import java.lang.Thread

from tweener import Tweener
from gamestate import GameState, GameStateManager
from ngram import NGramModel
from score_manager import ScoreManager, AsyncRemoteScoreManager
from sound_manager import SoundManager, Klaxon
import quests
import wordscore

import userdir

from paulscode.sound import SoundSystem
from paulscode.sound import SoundSystemConfig
from paulscode.sound.libraries import LibraryLWJGLOpenAL
from paulscode.sound.codecs import CodecWav
from paulscode.sound.codecs import CodecJOrbis

client_version = 101

specialchance = 24
ngramlength = 7
remote_baseurl = "http://api.lexcavator.com/v1"

letterfreq = ''.join([
	'j' * 5,
	'q' * 5,
	'x' * 7,
	'z' * 9,
	'w' * 10,
	'k' * 12,
	'v' * 14,
	'f' * 16,
	'y' * 18,
	'b' * 20,
	'h' * 24,
	'g' * 27,
	'm' * 28,
	'p' * 29,
	'u' * 32,
	'd' * 32,
	'c' * 39,
	'l' * 51,
	't' * 63,
	'o' * 64,
	'n' * 65,
	'r' * 68,
	'a' * 75,
	'i' * 88,
	's' * 93,
	'e' * 110
])

powerup_types = {
	'shaft': {'ch':'!', 'desc':'Sink shaft'},
	'excavate': {'ch':'>', 'desc':'Excavate current row'},
	'explosive': {'ch':'*', 'desc':'Clear 5x5 area'},
	'alchemy': {'ch':'?', 'desc':'Change all letters'},
	'erosion': {'ch':'_', 'desc':'Destroy random tiles'}
}

arcade_pps = {
	'easy': [1.00, 1.50, 2.00, 2.75, 3.50, 4.50, 5.75, 7.00, 8.50, 10.5, 13.0],
	'norm': [2.00, 3.00, 4.00, 5.25, 6.50, 7.75, 9.00, 10.5, 12.0, 15.0, 20.0],
	'hard': [3.00, 4.50, 6.00, 8.50, 10.0, 12.0, 14.0, 16.0, 18.0, 20.0, 23.0]
}

# will be loaded by LoadResourcesState
ngram_model = None
checker = None
score_manager = None
quest_manager = None
remote_manager = None
sound_manager = None

lwidth = 160
lheight = 240
game_scale = 2 # default, set dynamically in setup()

def lmouseX():
	return mouseX / game_scale

def lmouseY():
	return mouseY / game_scale

class WordChecker(object):
	def __init__(self):
		self.words = set()
	def feed(self, word):
		self.words.add(word)
	def is_word(self, word):
		return word in self.words

class LetterSprite(object):
	def __init__(self, x, y, bgcolor=(64, 64, 255), fgcolor=(255,255,255,255), ch='@', spsize=8):
		self.x = x
		self.y = y
		self.bgcolor = bgcolor
		self.bgalpha = 255
		self.fgcolor = fgcolor
		self.ch = ch
		self.xspeed = 0
		self.yspeed = 0
		self.accel = 0.06
		self.decel = 0.06
		self.maxxspd = 0.09
		self.maxyspd = 1
		self.xsave = 0
		self.ysave = 0
		self.xrep = 0
		self.yrep = 0
		self.gravity = 0.00075
		self.right = False
		self.left = False
		self.jump = False
		self.jump_force = 0.24
		self.spsize = spsize

	def update(self, playfield=None, delta=33):
		mult = 0
		if self.right:
			mult = 1
		elif self.left:
			mult = -1
		if self.right or self.left:
			self.xspeed += (self.accel * delta) * mult
			if abs(self.xspeed) > abs(self.maxxspd):
				self.xspeed = self.maxxspd * mult
		else:
			if self.xspeed > 0:
				self.xspeed -= (self.decel * delta)
				if self.xspeed < 0:
					self.xspeed = 0
			elif self.xspeed < 0:
				self.xspeed += (self.decel * delta)
				if self.xspeed > 0:
					self.xspeed = 0

		self.yspeed += (self.gravity * delta)

		if self.jump and playfield is not None:
			if (playfield.char_at_coord(self.x, self.y + self.spsize) != ' ') or \
					(playfield.char_at_coord(self.x+self.spsize-1, self.y + self.spsize) != ' '):
				sound_manager.play('blipblipblip.wav')
				self.yspeed = -self.jump_force
				self.jump = False

		self.xrep = 0
		self.yrep = 0
		self.xrep += int(abs(self.xspeed * delta))
		self.yrep += int(abs(self.yspeed * delta))
		self.xsave += abs(self.xspeed * delta) - int(abs(self.xspeed * delta))
		self.ysave += abs(self.yspeed * delta) - int(abs(self.yspeed * delta))
		signx = -1 if self.xspeed < 0 else 1
		signy = -1 if self.yspeed < 0 else 1
		offsetx = 0 if self.xspeed < 0 else self.spsize-1
		offsety = 0 if self.yspeed < 0 else self.spsize-1

		if self.xsave >= 1:
			self.xsave -= 1
			self.xrep += 1
		if self.ysave >= 1:
			self.ysave -= 1
			self.yrep += 1

		while self.yrep > 0:
			self.yrep -= 1
			if playfield is None or (playfield.char_at_coord(self.x, self.y+offsety+signy) == ' ' and \
					playfield.char_at_coord(self.x + self.spsize-1, self.y+offsety+signy) == ' '):
				self.y += signy
			else:
				self.jump = False
				self.yspeed = 0

		while self.xrep > 0:
			self.xrep -= 1
			if playfield is None or (playfield.char_at_coord(self.x+offsetx+signx, self.y) == ' ' and \
					playfield.char_at_coord(self.x+offsetx+signx, self.y + self.spsize-1) == ' ' and \
					self.x+offsetx+signx > 0 and self.x+offsetx+signx < lwidth):
				self.x += signx
			else:
				self.xspeed = 0

	def collides_with(self, other):
		if (self.x+1 > other.x+self.spsize-1) or \
				(self.x+self.spsize-1 < other.x+1) or \
				(self.y+1 > other.y+self.spsize-1) or \
				(self.y+self.spsize-1 < other.y+1):
			return False
		sound_manager.play('beep.wav')
		return True

	def draw(self):
		noStroke()
		fill(self.bgcolor[0], self.bgcolor[1], self.bgcolor[2], self.bgalpha)
		textSize(self.spsize)
		rect(self.x-1, self.y-1, self.spsize+1, self.spsize+1)
		fill(*self.fgcolor)
		text(self.ch, self.x + (self.spsize / 2), self.y + (self.spsize/2) - 1)

class PowerupSprite(LetterSprite):
	pass

class LetterDrag(object):
	def __init__(self):
		self.positions = list()
		self.word = str()
	def track_position(self, x, y, ch):
		# if this position is the same as the last one, bail
		if len(self.positions) > 0:
			last = self.positions[-1]
			if x == last[0] and y == last[1]:
				return
			# only add positions that are at most one square away
			if abs(x - last[0]) > 1 or abs(y - last[1]) > 1:
				return
		# if we're back over the previously added letter, player is backing up, so
		# remove from positions and string
		if len(self.positions) > 1 and self.positions[-2] == (x, y):
			self.positions.pop(-1)
			self.word = self.word[:-1]
			return

		if (x, y) in self.positions: return
		if ch == ' ': return
		self.positions.append((x, y))
		self.word += ch
		sound_manager.play('tick.wav')

	def get_positions(self):
		return self.positions
	def get_word(self):
		return self.word
	def get_height(self):
		miny = min([y for (x, y) in self.positions])
		maxy = max([y for (x, y) in self.positions])
		return abs(maxy - miny) + 1
	def get_width(self):
		minx = min([x for (x, y) in self.positions])
		maxx = max([x for (x, y) in self.positions])
		return abs(maxx - minx) + 1

class MenuState(GameState):
	def __init__(self, yoff=0, last=None):
		self.yoff = yoff
		self.menu = list()
		self.selected = None
		self.last = last

	def add_item(self, item):
		self.menu.append(item)

	def back(self):
		self.manager.remove_state(self)
		self.manager.unmute(self.last)

	def draw(self):
		moused_over_something = False
		for item in self.menu:
			this_spr = item['spr']
			if this_spr.mouse_over():
				moused_over_something = True
				(this_spr.r, this_spr.g, this_spr.b) = (255, 255, 255)
				if item != self.selected:
					sound_manager.play('tick.wav')
				self.selected = item
				if 'text' in item:
					textFont(fonts['tiny'])
					textAlign(CENTER)
					fill(227, 114, 20)
					text(item['text'], lwidth/2, lheight - 8)
					textFont(fonts['normal'])
			else:
				(this_spr.r, this_spr.g, this_spr.b) = (104, 234, 161)
			this_spr.draw()
		if moused_over_something is False:
			self.selected = None

	def mousePressed(self):
		if self.selected is None: return
		if callable(self.selected['action']):
			self.selected['action']()
			sound_manager.play('wibble.wav')

class PlayfieldState(GameState):
	# get the character at this map position
	def char_at_pos(self, x, y):
		check = (y * self.xsize) + x
		if check >= 0 and check < len(self.playfield):
			return self.playfield[(y * self.xsize) + x]
		else:
			return ' '

	def special_at_pos(self, x, y):
		check = (y * self.xsize) + x
		if check < len(self.specialmap):
			return self.specialmap[(y * self.xsize) + x]
		else:
			return 0

	def pos_at_coord(self, x, y):
		return (int(x/float(self.rectxsize)), int(y/float(self.rectysize)))

	# get the character at these map coordinates (pixels)
	def char_at_coord(self, x, y):
		ch = self.char_at_pos(int(x/float(self.rectxsize)), 
				int(y/float(self.rectysize)))
		return ch

	def char_at_screen(self, x, y):
		pos = self.screen_to_world(x, y)
		return self.char_at_coord(pos[0], pos[1])

	def pos_at_screen(self, x, y):
		pos = self.screen_to_world(x, y)
		return self.pos_at_coord(pos[0], pos[1])

	def screen_to_world(self, x, y):
		return (x, y - self.get_yoffset())

	# set the character at this position
	def set_at_pos(self, x, y, ch):
		check = (y * self.xsize) + x
		if check > len(self.playfield):
			self.playfield.extend(' ' * (check - len(self.playfield)))
		self.playfield[check] = ch

	def extend_playfield(self):
		extendrows = random.randrange(10, 20)
		#print "extending by %d rows" % extendrows
		gyoff = len(self.playfield) / self.xsize
		for i in range(extendrows):
			self.playfield.extend([' ' for j in range(self.xsize)])
			self.specialmap.extend([1 if random.randrange(specialchance)==0 else 0 \
					for i in range(self.xsize)])
		#print "now has %d rows" % (len(self.playfield) / self.xsize)
		# populate with random markov chains
		for i in range(6):
			for j in range(extendrows * self.xsize):
				x = random.randrange(self.xsize)
				y = gyoff + random.randrange(extendrows / 2)
				if self.char_at_pos(x, y) == ' ':
					break
			randommod = random.choice(self.mod)
			s = randommod.generate()
			#print "generated: %s from mod %d" % (s, randommod.n)
			for ch in s:
				self.set_at_pos(x, y, ch)
				#print "ch %s at %d, %d" % (ch, x, y)
				positions = [(-1,0),(1,0),(-1,1),(0,1),(0,1),(0,1),(0,1),(0,1),(1,1)]
				random.shuffle(positions)
				found = False
				for pos in positions:
					ny = y + pos[1]
					nx = x + pos[0]
					if nx >= self.xsize: nx = self.xsize - 1
					if nx < 0: nx = 0
					if self.char_at_pos(nx, ny) == ' ' and \
							(ny * self.xsize) + nx < len(self.playfield):
						x = nx
						y = ny
						found = True
						break
				if found is False:
					break

		# find empty spots, fill with markov chains
		for x in range(self.xsize):
			for y in range(gyoff, gyoff+extendrows):
				found = False
				if self.char_at_pos(x, y) == ' ':
					for mod in reversed(self.mod):
						upstr = ''.join([self.char_at_pos(x, (y-mod.n)+ny) for ny in range(mod.n)])
						nextc = mod.next_for(upstr)
						#print "found %s for mod.n %d nextc %s" % (upstr, mod.n, nextc)
						if nextc:
							#print "generated %s" % nextc
							self.set_at_pos(x, y, nextc[-1])
							found = True
							break
					if found is False:
						#print "nope, random letter"
						self.set_at_pos(x, y, random.choice(letterfreq))

	def __init__(self, xsize, ysize, checker, ngram_model):
		self.checker = checker
		self.mod = ngram_model
		self.playfield = array.array('c')
		self.specialmap = list()
		self.xsize = xsize
		self.ysize = ysize
		self.rectxsize = lwidth / self.xsize
		self.rectysize = lheight / self.ysize

		self.extend_playfield()

		for i in range(self.xsize):
			self.set_at_pos(i, 0, ' ')

		self.player = LetterSprite(0, 0)
		self.specials = list()
		self.particles = list()

		self.bezel = 4

		self.keys_down = set()
		self.drag = None

		self.words_formed = list()
		self.game_over = False
		self.paused = False

		self.player_deepest = 0

		self.words_formed = list()
		self.letters_used = 0
		self.start_time = millis()


	# should the display be offset by some number of pixels? override in
	# subclass
	def get_yoffset(self):
		return -self.player.y + (lheight * 0.5)

	# any code that needs to be done before drawing happens; override in
	# subclass
	def pre_draw(self):
		pass

	def update(self, delta):
		self.player.update(self, delta)
		for sp in self.specials:
			if not(self.paused):
				sp.update(self, delta)
		for sp in self.particles:
			if not(self.paused):
				sp.update(None, delta)

	def draw(self):

		# which positions are part of the current drag?
		dragged = []
		if self.drag:
			dragged = self.drag.get_positions()

		self.pre_draw()

		if self.player.y > self.player_deepest:
			self.player_deepest = self.player.y

		yoffset = self.get_yoffset()

		imageMode(CORNER)
		bgheight = images['background'].height
		bgpos = (-lheight * 0.5 + yoffset * 0.5) % (bgheight * -1)

		image(images['background'], 0, bgpos)
		if bgheight + bgpos < lheight:
			image(images['background'], 0, (bgpos+bgheight))

		pushMatrix()
		translate(0, yoffset)

		# mouseover letter only works inside bezel (to make diagonal
		# selection easier)
		mousepos = self.pos_at_screen(lmouseX(), lmouseY())
		mousecoord = self.screen_to_world(lmouseX(), lmouseY())
		mousexoff = mousecoord[0] % self.rectxsize
		mouseyoff = mousecoord[1] % self.rectysize
		if mousexoff < self.bezel or mousexoff > self.rectxsize - self.bezel \
				or mouseyoff < self.bezel or mouseyoff > self.rectysize - self.bezel:
			mousepos = None

		# draw letters
		textAlign(CENTER, CENTER)
		imageMode(CORNER)
		noTint()
		textSize(8)
		for x in range(self.xsize):
			for y in range(len(self.playfield) / self.xsize):
				ch = self.char_at_pos(x, y)
				if ch == ' ': continue
				pushMatrix()
				translate(x * self.rectxsize, y * self.rectysize)
				if not(self.paused) and mousepos and x == mousepos[0] and y == mousepos[1]:
					pushMatrix()
					translate(self.rectxsize / 2, self.rectysize / 2)
					imageMode(CENTER)
					tint(255, 192 + (sin(frameCount * 0.2) * 64))
					image(images['selector1'], 0, 0)
					popMatrix()
				noTint()
				imageMode(CORNER)
				if (x, y) in dragged:
					fill(113, 53, 3)
					# wiggle dragged special tiles
					if self.special_at_pos(x, y):
						translate(random.randrange(-2, 2), random.randrange(-2, 2))
					image(images['orangeblock'], 0, 0)
				elif self.special_at_pos(x, y):
					image(images['greenblock'], 0, 0)
					fill(255, 255, 255)
				else:
					image(images['cyanblock'], 0, 0)
					fill(2, 68, 68)
				text(ch, self.rectxsize / 2, self.rectysize / 2)
				popMatrix()
		self.player.draw()
		for sp in self.specials:
			sp.draw()
			if self.player.collides_with(sp):
				self.execute_effect(sp)
		for sp in self.particles:
			sp.draw()

		if self.drag is not None:
			self.draw_drag()
				
		if not(self.paused) and mousepos and mousePressed and self.drag is not None:
			ch = self.char_at_screen(lmouseX(), lmouseY())
			self.drag.track_position(mousepos[0], mousepos[1], ch)

		player_screeny = screenY(0, self.player.y)
		if player_screeny + 8 < 0 and self.game_over is False:
			self.game_over = True
			self.paused = True
			self.end_game()

		popMatrix()

		# keep extending board
		if len(self.playfield) < (self.xsize * (self.player.y / self.rectysize)) +\
				self.xsize * 50:
			self.extend_playfield()

		textFont(fonts['tiny'])
		textAlign(LEFT)
		textFont(fonts['normal'])

	def end_game(self):
		elapsed_millis = millis() - self.start_time
		self.manager.add_state(GameOverState(to_remove=self,
				elapsed_seconds=elapsed_millis / 1000,
				mode="none", words_formed=self.words_formed,
				letters_used=self.letters_used))

	def execute_effect(self, sp):
		#print "in execute_effect!"
		self.specials.remove(sp)
		player_x, player_y = self.pos_at_coord(self.player.x + 4, self.player.y + 4)
		if sp.effect == 'shaft':
			# remove column beneath player
			positions = [(player_x, player_y + i) for i in range(1, 5)]
			self.remove_positions(positions)
		elif sp.effect == 'excavate':
			# remove player's current row
			positions = [(i, player_y) for i in range(self.xsize)]
			self.remove_positions(positions)
		elif sp.effect == 'explosive':
			# remove 5x5 square surrounding player
			positions = list()
			for ox in range(-2, 3):
				for oy in range(-2, 3):
					positions.append((player_x + ox, player_y + oy))
			self.remove_positions(positions)
		elif sp.effect == 'alchemy':
			# change every tile currently on the board
			topx, topy = self.pos_at_screen(0, 0)
			for ox in range(topx, topx + self.xsize):
				for oy in range(topy, topy + self.ysize + (self.ysize / 2)):
					if self.char_at_pos(ox, oy) != ' ':
						self.set_at_pos(ox, oy, random.choice(letterfreq))
		elif sp.effect == 'erosion':
			# delete tiles at random (favoring common letters)
			topx, topy = self.pos_at_screen(0, 0)
			positions = list()
			for ox in range(topx, topx + self.xsize):
				for oy in range(topy, topy + self.ysize + (self.ysize / 2)):
					if self.char_at_pos(ox, oy) in 'esiarnot' \
							and random.randrange(10) <= 2:
						positions.append((ox, oy))
					if self.char_at_pos(ox, oy-1) == ' ' \
							and random.randrange(10) <= 1:
						positions.append((ox, oy))
			self.remove_positions(positions)

	def draw_drag(self):
		dragged = self.drag.get_positions()
		if len(dragged) > 1:
			for i in range(len(dragged) - 1):
				this = dragged[i]
				next = dragged[i+1]
				diffx = this[0] - next[0]
				diffy = this[1] - next[1]
				# north
				pushMatrix()
				imageMode(CENTER)
				noTint()
				translate(next[0] * self.rectxsize + (self.rectxsize/2),
						next[1] * self.rectysize + (self.rectysize/2))
				if diffx == 0 and diffy == 1:
					rotate(0)
					image(images['narrow_mask'], 1, 9)
					image(images['narrow'], 0, 8)
				elif diffx == -1 and diffy == 0:
					rotate(PI/2)
					image(images['narrow_mask'], 1, 9)
					image(images['narrow'], 0, 8)
				elif diffx == 0 and diffy == -1:
					rotate(PI)
					image(images['narrow_mask'], 1, 9)
					image(images['narrow'], 0, 8)
				elif diffx == 1 and diffy == 0:
					rotate(PI * 1.5)
					image(images['narrow_mask'], 1, 9)
					image(images['narrow'], 0, 8)
				elif diffx == -1 and diffy == 1:
					rotate(0)
					image(images['nwarrow_mask'], -7, 9)
					image(images['nwarrow'], -8, 8)
				elif diffx == -1 and diffy == -1:
					rotate(PI/2)
					image(images['nwarrow_mask'], -7, 9)
					image(images['nwarrow'], -8, 8)
				elif diffx == 1 and diffy == -1:
					rotate(PI)
					image(images['nwarrow_mask'], -7, 9)
					image(images['nwarrow'], -8, 8)
				elif diffx == 1 and diffy == 1:
					rotate(PI * 1.5)
					image(images['nwarrow_mask'], -7, 9)
					image(images['nwarrow'], -8, 8)
				popMatrix()

	def pause(self):
		self.paused = True
		self.manager.add_state(PauseHelpState(callback=self.unpause,
				to_remove=self))

	def keyPressed(self):
		if self.paused: return
		if (key == CODED and keyCode == RIGHT) or key == ord('d'):
			self.player.right = True
		elif (key == CODED and keyCode == LEFT) or key == ord('a'):
			self.player.left = True
		elif (key == CODED and keyCode == UP) or key == ord('w') \
				or key == ord(' '):
			if UP not in self.keys_down:
				self.player.jump = True
			self.keys_down.add(UP)
			self.jumped()
		elif key == 27:
			self.pause()

	def jumped(self):
		pass

	def keyReleased(self):
		if self.paused: return
		if (key == CODED and keyCode == RIGHT) or key == ord('d'):
			self.player.right = False
		elif (key == CODED and keyCode == LEFT) or key == ord('a'):
			self.player.left = False
		elif (key == CODED and keyCode == UP) or key == ord('w') \
				or key == ord(' '):
			self.player.jump = False
			if UP in self.keys_down:
			 	self.keys_down.remove(UP)

	def mousePressed(self):
		if self.paused: return
		self.drag = LetterDrag()

	def mouseReleased(self):
		if self.paused: return
		if self.drag is None: return
		word = self.drag.get_word()
		if self.checker.is_word(word):
			self.remove_dragged_word()
			(score, mult) = self.score_word(word)
			#print word, score, mult
			self.word_formed(word, score, mult, self.drag)
		else:
			sound_manager.play('thud.wav')
		self.drag = None

	# what should happen when a word is formed?
	def word_formed(self, word, score, mult, drag):
		self.words_formed.append({"word": word, "score": score})
		self.letters_used += len(word)

	def unpause(self):
		self.paused = False

	def score_word(self, word):
		return wordscore.score_word(word)

	def remove_positions(self, positions):
		sound_manager.play('squish.wav')
		for px, py in positions:
			if self.char_at_pos(px, py) == ' ':
				continue
			if self.special_at_pos(px, py):
				# add a power-up
				sp = PowerupSprite(px*self.rectxsize + (self.rectxsize/4),
						py*self.rectysize + (self.rectysize/4), ch='!')
				self.specials.append(sp)
				#sp.effect = random.choice(powerup_types.keys())
				sp.effect = random.choice(['shaft', 'excavate', 'explosive', 'alchemy',
						'erosion'])
				sp.ch = powerup_types[sp.effect]['ch']
				T.addTween(sp, x=random.randrange(-15, 15)*0.01, tweenTime=500,
						tweenType=T.LINEAR)
			self.set_at_pos(px, py, ' ')
			for i in range(random.randrange(2,6)):
				sp = LetterSprite(px*self.rectxsize + (self.rectxsize/4),
						py*self.rectysize + (self.rectysize/4),
						bgcolor=(249, 173, 110),
						ch=' ',
						spsize=random.randrange(6,16))
				sp.yspeed = -1 * random.randrange(100) * 0.002
				dir = random.randrange(2)
				sp.accel = sp.decel = random.randrange(5) * 0.001
				if dir == 0: sp.right = True
				if dir == 1: sp.left = True
				self.particles.append(sp)
				T.addTween(sp, bgalpha=-255, tweenTime=random.randrange(500,1000),
						tweenType=T.IN_OUT_QUAD,
						onCompleteFunction=lambda s=sp: self.remove_particle(s))

	def remove_particle(self, sp):
		self.particles.remove(sp)

	def remove_dragged_word(self):
		self.remove_positions(self.drag.get_positions())

	def state_added(self): pass
	def state_removed(self): pass

class TutorialModePlayfieldState(PlayfieldState):

	# override in subclasses to add behavior when the state is added
	def state_added(self):
		self.tutorial_overlay = TutorialOverlayState()
		T.delay(1000, lambda: self.manager.add_state(self.tutorial_overlay))

	def state_removed(self):
		if self.tutorial_overlay in self.manager.states:
			self.manager.remove_state(self.tutorial_overlay)

	def word_formed(self, word, score, mult, drag):
		self.tutorial_overlay.has_formed_word()
		super(TutorialModePlayfieldState, self).word_formed(word, score, mult,
				drag)

	def jumped(self):
		self.tutorial_overlay.has_jumped()

	def execute_effect(self, sp):
		self.tutorial_overlay.has_collected_powerup()
		super(TutorialModePlayfieldState, self).execute_effect(sp)

class TutorialOverlayState(GameState):

	def __init__(self):
		self.overlay_text = [
			"Welcome to Lexcavator! You control @. Move with arrow keys or WASD. UP (or <W>) jumps. Try jumping now!",
			"Your goal is to move @ as deep as possible. Click and drag the mouse to form words from tiles, removing them from the board. Words must be at least 3 letters long. Try it now!",
			"Form a word with a green tile to release a power-up. Collect the power-up to activate it. Try collecting a power-up now!",
			"Hit <ESC> to pause the game and see a list of power-ups. In Quest Mode, <ESC> shows quest progress. Try it now!",
			"Congrats! Now you know how to lexcavate. Hit <T> to hide this message and continue practicing in tutorial mode. Hit <ESC> to return to the main menu to select another mode."]
		self.overlays = [textwrap.wrap(x, 24) for x in self.overlay_text]
		self.switch_time = 7000
		self.idx = 0
		self.jumped = False
		self.formed_word = False
		self.collected_powerup = False
		self.hit_escape = False

	def keyPressed(self):
		if key == 27:
			self.hit_escape = True
		elif key == ord('t'):
			self.manager.mute(self)

	def has_jumped(self):
		self.jumped = True

	def has_formed_word(self):
		self.formed_word = True

	def has_collected_powerup(self):
		self.collected_powerup = True

	def draw(self):

		if self.jumped is False:
			self.idx = 0
		elif self.formed_word is False:
			self.idx = 1
		elif self.collected_powerup is False:
			self.idx = 2
		elif self.hit_escape is False:
			self.idx = 3
		else:
			self.idx = 4

		fill(0, 128)
		rect(8, 8, lwidth-16, 3 + (len(self.overlays[self.idx]) * 8))
		textFont(fonts['tiny'])
		textAlign(LEFT)
		fill(255)
		yoff = 16
		for line in self.overlays[self.idx]:
			text(line, 11, yoff)
			yoff += 8
		textFont(fonts['normal'])

class QuestModePlayfieldState(PlayfieldState):

	def __init__(self, xsize, ysize, checker, ngram_model):

		self.mode_complete = quest_manager.all_complete()
		self.depth = 0.0

		super(QuestModePlayfieldState, self).__init__(xsize, ysize, checker,
				ngram_model)

	def pre_draw(self):
		self.depth = self.player.y / float(self.rectysize * 26)

	def state_added(self):
		self.overlay = QuestModeOverlayState()
		self.manager.add_state(self.overlay)
		sound_manager.set_music('whatwevegot.ogg')
		self.pause()
		super(QuestModePlayfieldState, self).state_added()

	def word_formed(self, word, score, mult, drag):
		super(QuestModePlayfieldState, self).word_formed(word, score, mult, drag)
		lev_before = quest_manager.get_current_level()
		quest_manager.word_formed(word, score, mult, drag,
				callback=self.overlay.add_quest_notification)
		self.post_quest_action(lev_before)

	def execute_effect(self, sp):
		super(QuestModePlayfieldState, self).execute_effect(sp)
		lev_before = quest_manager.get_current_level()
		quest_manager.power_up_collected(sp.effect,
				callback=self.overlay.add_quest_notification)
		self.post_quest_action(lev_before)

	def post_quest_action(self, lev_before):
		lev_after = quest_manager.get_current_level()
		if quest_manager.all_complete() and self.mode_complete is False:
			self.mode_complete = True
			self.overlay.all_complete()
		if lev_after > lev_before:
			self.overlay.level_unlocked(lev_after)

	def pause(self):
		self.paused = True
		elapsed_millis = millis() - self.start_time
		self.manager.add_state_above(self,
				QuestModeProgressState(callback=self.unpause,
				to_remove=self, words_formed=self.words_formed, depth=self.depth,
				letters_used=self.letters_used, elapsed_seconds=elapsed_millis/1000))

class QuestModeOverlayState(GameState):

	def __init__(self):
		self.notifs = list()
		self.notif_queue = list()
		self.notif_yoff = 0
		self.yoff_delta = 17
		self.last_pop = 0
		self.pop_interval = 500
		self.scrolling_in = 0

	def draw(self):
		textFont(fonts['tiny'])
		textAlign(LEFT)
		for notif in self.notifs:
			notif.draw()
		textFont(fonts['normal'])
		if millis() > self.last_pop + self.pop_interval:
			self.last_pop = millis()
			self.pop_notif_queue()

	def add_quest_notification(self, quest, ntype):
		notif = QuestNotification(quest, ntype)
		self.add_notification(notif)

	def add_notification(self, notif):
		self.notif_queue.append(notif)
		if self.scrolling_in == 0:
			self.last_pop = millis()
			self.pop_notif_queue()

	def pop_notif_queue(self):
		if len(self.notif_queue) == 0: return
		notif = self.notif_queue.pop(0)
		self.notifs.append(notif)
		T.addTween(notif, a=-255, tweenDelay=1000, tweenTime=5000,
				tweenType=T.IN_OUT_QUAD,
				onCompleteFunction=lambda: self.remove_notification(notif))
		self.scrolling_in += 1
		T.delay(500, self.decrement_scrolling_in)
		for n in self.notifs:
			T.addTween(n, yoff=-float(self.yoff_delta), tweenTime=500,
					tweenType=T.IN_OUT_QUAD)

	def decrement_scrolling_in(self): self.scrolling_in -= 1

	def level_unlocked(self, lev):
		notif = QuestMetaNotification(lines=[
			"New quests unlocked!",
			"Hit <ESC> to view"
		])
		self.add_notification(notif)

	def all_complete(self):
		notif = QuestMetaNotification(lines=[
			"Quest mode complete!",
			"Congratulations!"])
		self.add_notification(notif)

	def remove_notification(self, notif):
		self.notifs.remove(notif)

class QuestNotification(object):
	def __init__(self, quest, ntype, yoff=lheight):
		self.quest = quest
		self.ntype = ntype
		self.progress = quest.progress
		self.goal = quest.goal
		self.a = 255
		self.yoff = yoff
	def draw(self):
		pushMatrix()
		translate(0, self.yoff)
		fill(0, self.a * 0.5)
		rect(4, 0, lwidth-6, 16)
		fill(255, self.a)
		text(self.quest.title, 6, 7)
		if type(self.ntype) is quests.QuestProgress:
			text("%d out of %d" % (self.progress,
					self.goal), 6, 14)
		elif type(self.ntype) is quests.StreakBroken:
			text("Streak broken! Count reset", 6, 14)
		elif type(self.ntype) is quests.QuestComplete:
			text("Quest complete!", 6, 14)
		popMatrix()

class QuestMetaNotification(object):
	def __init__(self, lines, yoff=lheight):
		self.lines = lines
		self.yoff = yoff
		self.a = 255
	def draw(self):
		pushMatrix()
		translate(0, self.yoff)
		fill(0, self.a * 0.5)
		rect(4, 0, lwidth-6, 16)
		fill(244, 159, 82, self.a)
		text(self.lines[0], 6, 7)
		fill(255, self.a)
		text(self.lines[1], 6, 14)
		popMatrix()

class QuestModeProgressState(MenuState):

	def __init__(self, callback, to_remove, words_formed=[], letters_used=0,
			elapsed_seconds=0, depth=0):
		self.callback = callback
		self.to_remove = to_remove
		self.init_millis = millis()
		self.words_formed_count = len(words_formed)
		self.letters_used = letters_used
		self.elapsed_seconds = elapsed_seconds
		self.depth = depth
		self.ended = False
		self.current_level = quest_manager.get_current_level() or 0
		self.image_yoff = -100
		self.tween_complete = False
		super(QuestModeProgressState, self).__init__(self)

	def state_added(self):

		def next_f():
			self.current_level += 1
			if self.current_level >= len(quest_manager.quests):
				self.current_level = 0

		def prev_f():
			self.current_level -= 1
			if self.current_level < 0:
				self.current_level = len(quest_manager.quests) - 1

		nextlevspr = ShadowStringSprite(">", lwidth/2 + 40, 108, align=CENTER)
		self.add_item({'spr': nextlevspr, 'action': next_f})
		prevlevspr = ShadowStringSprite("<", lwidth/2 - 40, 108, align=CENTER)
		self.add_item({'spr': prevlevspr, 'action': prev_f})
		def play_f():
			self.manager.remove_state(self)
			self.callback()
		self.add_item({'spr': ShadowStringSprite("Play", lwidth / 2,
				lheight - 8, align=CENTER), 'action': play_f})
		def menu_f():
			self.ended = True
			self.update_stats()
			self.manager.add_state(BlockTransitionState(self, TitleScreenState()))
		self.add_item({'spr': ShadowStringSprite("Menu", lwidth / 5,
				lheight - 8, align=CENTER), 'action': menu_f})
		def help_f():
			self.manager.remove_state(self)
			self.manager.add_state(QuestModeHelpState(self.callback))
		self.add_item({'spr': ShadowStringSprite("Help", lwidth * 0.8,
				lheight - 8, align=CENTER), 'action': help_f})

		T.addTween(self, image_yoff=int(lheight * 0.2)+100,
			tweenType=T.IN_OUT_QUAD, tweenTime=500, onCompleteFunction=self.tweened)

	def tweened(self):
		self.tween_complete = True

	def draw(self):
		fill(0, 192)
		noStroke()
		rect(0, 0, lwidth, lheight)
		fill(0, 32)
		for i in range(0, lheight, 2):
			rect(0, float(i), lwidth, 1)

		imageMode(CENTER)
		tint(0, 128)
		image(images['quest_progress_shadow'], lwidth/2 + 2, self.image_yoff + 2)
		noTint()
		image(images['quest_progress'], lwidth/2, self.image_yoff)

		if self.tween_complete:

			fill(255)
			textSize(8)
			textAlign(CENTER)
			text("Level %d" % (self.current_level+1), lwidth/2, 108)

			textSize(8)
			textFont(fonts['tiny'])
			textAlign(LEFT)
			yoff = 124
			
			for quest in quest_manager.quests[self.current_level]:
				imageMode(CORNER)
				if quest.is_completed():
					image(images['checked'], 7, yoff-7)
				else:
					image(images['checkbox'], 8, yoff-6)
				fill(244, 159, 82)
				text(quest.title, 20, yoff)
				fill(255)
				if quest.goal > 1:
					text("%d/%d" % (quest.progress, quest.goal),
							24+textWidth(quest.title), yoff)
				yoff += 7
				if quest_manager.get_current_level() is not None and \
						(self.current_level > quest_manager.get_current_level()):
					desc = "LOCKED!"
				else:
					desc = quest.desc
				for line in textwrap.wrap(desc, 24):
					text(line, 20, yoff)
					yoff += 7
				yoff += 4

			if quest_manager.get_current_level() is not None and \
					(self.current_level > quest_manager.get_current_level()):
				textAlign(CENTER)
				text("Complete all quests in", lwidth/2, lheight-31)
				text("level %d to unlock!" % (self.current_level), lwidth/2,
						lheight-24)

			textFont(fonts['normal'])
			super(QuestModeProgressState, self).draw()

	def keyPressed(self):
		if self.ended: return
		if key == CODED: return
		if millis() > self.init_millis + 250:
			if key == ord('q'):
				self.ended = True
				self.update_stats()
				self.manager.add_state(BlockTransitionState(self, TitleScreenState()))
			else:
				self.manager.remove_state(self)
				self.callback()

	def update_stats(self):
		score_manager.update_stats("quest",
			words_formed=self.words_formed_count,
			letters_used=self.letters_used,
			depth=self.depth,
			seconds_elapsed=self.elapsed_seconds)

	def state_removed(self):
		if self.ended:
			self.manager.remove_state(self.to_remove)

class QuestModeHelpState(MenuState):

	def __init__(self, callback):
		self.callback = callback
		super(QuestModeHelpState, self).__init__(self)

	def state_added(self):
		self.help_text = """
Enjoy these special
challenges designed to
test your lexcavation
skills!

> This mode has no time
  limit. Play at your own
  pace!

> You can complete quests
  in any order, but you
  must complete all quests
  in one level to advance
  to the next.

> Hit <ESC> to see your
  current progress or
  quit to the main menu.

> Don't worry: Your
  progress will be
  saved automatically.

Good luck!"""
		def dismiss_f():
			self.manager.remove_state(self)
		self.add_item({'spr': ShadowStringSprite("Got it!", lwidth/2,
				lheight - 8, align=CENTER), 'action': dismiss_f})

	def draw(self):

		fill(0, 192)
		noStroke()
		rect(0, 0, lwidth, lheight)
		fill(0, 32)
		for i in range(0, lheight, 2):
			rect(0, float(i), lwidth, 1)

		fill(255)
		textSize(8)
		textAlign(CENTER)
		text("Quest Mode Help", lwidth/2, 16)

		textSize(8)
		textFont(fonts['tiny'])
		textAlign(LEFT)
		yoff = 26
		for line in self.help_text.split("\n"):
			text(line, 10, yoff)
			yoff += 7

		textFont(fonts['normal'])
		super(QuestModeHelpState, self).draw()

	def state_removed(self):
		self.callback()

class ChallengeModePlayfieldState(PlayfieldState):

	def challenge_rate_for_level(self, lev):
		if lev < len(arcade_pps[self.difficulty]):
			return arcade_pps[self.difficulty][lev] * 0.001
		else:
			return arcade_pps[self.difficulty][-1] * 0.001

	def __init__(self, xsize, ysize, checker, ngram_model, difficulty='easy'):

		assert difficulty in arcade_pps, "unknown difficulty level %s" % difficulty
		self.challenge_offset = 0
		self.difficulty = difficulty
		self.challenge_level = 1
		self.challenge_rate = self.challenge_rate_for_level(self.challenge_level)

		self.hiscore = 0
		hiscores = score_manager.get_high_scores("challenge_"+difficulty)
		if len(hiscores) > 0:
			self.hiscore = hiscores[0].score

		self.bestdepth = 0
		bestdepths = score_manager.get_best_depths("challenge_"+difficulty)
		if len(bestdepths) > 0:
			self.bestdepth = bestdepths[0].depth

		self.klaxon = Klaxon(sound_manager)

		super(ChallengeModePlayfieldState, self).__init__(xsize, ysize, checker,
				ngram_model)

	def update(self, delta):

		super(ChallengeModePlayfieldState, self).update(delta)

		if self.player.y > self.player_deepest:
			diff = self.player.y - self.player_deepest
			self.challenge_offset += diff * 0.025 * delta

		if not(self.paused):
			self.challenge_offset -= (self.challenge_rate * delta)
			if self.challenge_offset > lheight * 0.2:
				self.challenge_offset = lheight * 0.2

		self.info_overlay.set_depth(self.player.y / float(self.rectysize * 26))

		pscry = self.player.y + self.get_yoffset()
		if pscry < int(lheight * 0.1) and self.klaxon.sounding is False and \
				self.game_over is False:
			self.klaxon.start()
		elif pscry >= int(lheight * 0.1) and self.klaxon.sounding:
			self.klaxon.stop()
		self.klaxon.update()

	def get_yoffset(self):
		return -self.player.y + (lheight * 0.5) + int(self.challenge_offset)

	def state_added(self):
		self.info_overlay = InfoOverlayState(hiscore=self.hiscore,
				bestdepth=self.bestdepth)
		self.manager.add_state(self.info_overlay)
		sound_manager.set_music('gutfeeling.ogg')
		super(ChallengeModePlayfieldState, self).state_added()

	def word_formed(self, word, score, mult, drag):
		super(ChallengeModePlayfieldState, self).word_formed(word, score, mult,
				drag)
		self.info_overlay.word_formed(word, score, mult, lmouseX(),
				lmouseY())
		self.info_overlay.increment_score(score)
		if len(self.words_formed) % 8 == 0:
			self.challenge_level += 1
			self.challenge_rate = \
					self.challenge_rate_for_level(self.challenge_level)
			self.info_overlay.level_up(self.challenge_level)
			#print "level %d" % self.challenge_level

	def end_game(self):
		self.klaxon.stop()
		elapsed_millis = millis() - self.start_time
		self.manager.remove_state(self.info_overlay)
		self.manager.add_state(GameOverState(to_remove=self,
				mode="challenge_" + self.difficulty,
				words_formed=self.words_formed,
				letters_used=self.letters_used,
				score=self.info_overlay.score,
				depth=self.info_overlay.depth,
				elapsed_seconds=elapsed_millis/1000))

	def state_removed(self):
		self.klaxon.stop()
		super(ChallengeModePlayfieldState, self).state_removed()

class InfoOverlayState(GameState):

	def __init__(self, hiscore=0, bestdepth=0.0):
		self.drawables = list()
		self.score = 0
		self.hiscore = hiscore
		self.depth = 0.0
		self.bestdepth = bestdepth

	def draw(self):
		fill(0, 127)
		rect(0, lheight-17, lwidth, 24)

		fill(255)
		textAlign(LEFT)
		textFont(fonts['tiny'])
		text("SCORE", 4, lheight-9)
		text("%07d" % self.score, 35, lheight-9)
		text("BEST", 4, lheight-2)
		text("%07d" % self.hiscore, 35, lheight-2)
		text("DEPTH", lwidth-70, lheight-9)
		text("%03.2fab" % self.depth, lwidth-38, lheight-9)
		text("BEST", lwidth-70, lheight-2)
		text("%03.2fab" % self.bestdepth, lwidth-38, lheight-2)
		textFont(fonts['normal'])

		for d in self.drawables:
			if hasattr(d, 'update'):
				d.update()
		for d in self.drawables:
			d.draw()

	def increment_score(self, amount):
		self.score += amount

	def set_depth(self, amount):
		self.depth = amount

	def level_up(self, lev):
		spr = ImageSprite(images['levelup'], lwidth/2, lheight + 100)
		spr2 = ImageSprite(images['levelup_shadow'], lwidth/2 - 2, lheight + 98,
			a=128)
		T.addTween(spr, y=(lheight + 200) * -1, tweenTime=2500,
				tweenType=T.OUT_QUAD,
				onCompleteFunction=lambda: self.remove_drawable(spr))
		T.addTween(spr2, y=(lheight + 200) * -1, tweenTime=2500,
				tweenType=T.OUT_QUAD,
				onCompleteFunction=lambda: self.remove_drawable(spr))
		self.add_drawable(spr2)
		self.add_drawable(spr)

	def add_multiplier_sprite(self, mult, x, y):
		if x > lwidth - 21:
			x = lwidth - 21 
		if x < 21:
			x = 21 
		spr2 = TintCycleImageSprite(images['bonus'], x, y, a=255, delay=100,
				tints=[(173, 201, 231), (113, 169, 231), (104, 234, 161),
				(249, 173, 110)], shadow=True)
		T.addTween(spr2, y=-100, tweenTime=2000, tweenType=T.OUT_QUAD)
		T.addTween(spr2, a=-255, tweenTime=1500, tweenDelay=500,
				tweenType=T.OUT_QUAD,
				onCompleteFunction=lambda: self.remove_drawable(spr2))
		self.add_drawable(spr2)

	def word_formed(self, word, score, mult, x, y):
		st = "%s: %d" % (word, score)
		textFont(fonts['tiny'])
		stwidth = textWidth(st)
		if x > lwidth - (stwidth * 0.5):
			x = lwidth - (stwidth * 0.5)
		elif x < (stwidth * 0.5):
			x = stwidth * 0.5
		spr = TinyShadowStringSprite(st, x, y, textsize=8, r=249, g=173, b=110,
				a=255)
		T.addTween(spr, y=-100, tweenTime=2000, tweenType=T.OUT_QUAD)
		T.addTween(spr, a=-255, tweenTime=1500, tweenDelay=500,
				onCompleteFunction=lambda: self.remove_drawable(spr))
		self.add_drawable(spr)
		if len(word) >= 5:
			T.delay(250, lambda: self.add_multiplier_sprite(mult, x, y))
		textFont(fonts['normal'])

	def add_drawable(self, d):
		self.drawables.append(d)

	def remove_drawable(self, d):
		if d in self.drawables:
			self.drawables.remove(d)

class TimeTrialPlayfieldState(PlayfieldState):
	def __init__(self, xsize, ysize, checker, ngram_model, time_limit=90):
		self.time_limit = time_limit
		self.hiscore = 0
		hiscores = score_manager.get_high_scores("timetrial_"+str(time_limit))
		if len(hiscores) > 0:
			self.hiscore = hiscores[0].score

		self.bestdepth = 0
		bestdepths = score_manager.get_best_depths("timetrial_"+str(time_limit))
		if len(bestdepths) > 0:
			self.bestdepth = bestdepths[0].depth

		super(TimeTrialPlayfieldState, self).__init__(xsize, ysize, checker,
				ngram_model)

	def state_added(self):
		self.timer_overlay = TimerOverlayState(self, self.time_limit, self.hiscore,
				self.bestdepth)
		self.manager.add_state(self.timer_overlay)
		sound_manager.set_music('gutfeeling.ogg')
		super(TimeTrialPlayfieldState, self).state_added()

	def pre_draw(self):
		self.timer_overlay.set_depth(self.player.y / float(self.rectysize * 26))

	def word_formed(self, word, score, mult, drag):
		self.timer_overlay.word_formed(word, score, mult, lmouseX(), lmouseY())
		self.timer_overlay.increment_score(score)
		super(TimeTrialPlayfieldState, self).word_formed(word, score, mult, drag)

	def times_up(self):
		self.game_over = True
		self.paused = True
		self.end_game()

	def end_game(self):
		elapsed_millis = millis() - self.start_time
		self.manager.remove_state(self.timer_overlay)
		self.manager.add_state(GameOverState(to_remove=self,
				mode="timetrial_" + str(self.time_limit),
				words_formed=self.words_formed,
				letters_used=self.letters_used,
				elapsed_seconds=self.time_limit,
				score=self.timer_overlay.score,
				depth=self.timer_overlay.depth))

	def time_warning(self):
		# when time is running low
		pass

	def state_removed(self):
		if self.timer_overlay in self.manager.states:
			self.manager.remove_state(self.timer_overlay)

def minutes_seconds(remaining):
	minutes = remaining / 60
	second_remainder = remaining % 60
	return (minutes, second_remainder)

class TimerOverlayState(InfoOverlayState):

	def __init__(self, playfield, time_limit, hiscore=0, bestdepth=0):
		self.playfield = playfield
		self.seconds = time_limit
		self.last_remaining = 0
		self.called_times_up = False
		self.warning = False
		super(TimerOverlayState, self).__init__(hiscore=hiscore,
				bestdepth=bestdepth)

	def state_added(self):
		self.timer_sprite = ShadowStringSprite('TIME REMAINING %d:%0d' % (0, 0),
				lwidth/2, 8, textsize=8, r=113, g=169, b=231, a=255, align=CENTER)
		self.drawables.append(self.timer_sprite)
		self.started = millis()
		self.klaxon = Klaxon(sound_manager)

	def draw(self):

		delta = millis() - self.started
		delta_seconds = int(delta / 1000)
		remaining = self.seconds - delta_seconds

		# play tick if eight or fewer seconds left
		#if remaining != self.last_remaining and remaining <= 8 and remaining > 0:
		#	sounds['tick'].play(0)
		self.last_remaining = remaining

		# red if only one eight of the time remains
		textSize(32)
		textAlign(CENTER)
		if self.warning is False and remaining <= self.seconds / 8:
			self.klaxon.start()
			self.timer_sprite.r, self.timer_sprite.g, self.timer_sprite.b = \
				255, 0, 0
			self.warning = True
		self.klaxon.update()

		# call callback if time's up
		if remaining <= 0:
			remaining = 0
			if self.called_times_up is False:
				self.called_times_up = True
				self.playfield.times_up()

		self.timer_sprite.content = 'REMAINING %d:%02d' % \
				minutes_seconds(remaining)

		fill(0, 127)
		rect(0, 0, lwidth, 12)
		super(TimerOverlayState, self).draw()

	def state_removed(self):
		self.klaxon.stop()

class PauseHelpState(GameState):
	def __init__(self, callback, to_remove):
		self.callback = callback
		self.init_millis = millis()
		self.to_remove = to_remove
		self.ended = False
		self.items = [
			{'ch': '!', 'title': 'Shaft', 'desc': 'Sink shaft in current column'},
			{'ch': '>', 'title': 'Cut', 'desc': 'Clear current row'},
			{'ch': '*', 'title': 'Blast', 'desc': 'Clear 5x5 area'},
			{'ch': '?', 'title': 'Alchemize', 'desc': 'Randomize all letters'},
			{'ch': '_', 'title': 'Erode', 'desc': 'Destroy random tiles'}]

	def draw(self):
		fill(3, 37, 74)
		rect(0, 0, lwidth, lheight)
		imageMode(CENTER)
		tint(0, 128)
		image(images['paused_shadow'], lwidth/2 - 3, lheight * 0.2 - 3)
		noTint()
		image(images['paused'], lwidth/2, lheight * 0.2)
		textAlign(CENTER)
		textSize(8)
		yoff = 96
		fill(255)
		text("Power-up Guide", lwidth/2, yoff)
		yoff += 15

		textAlign(LEFT)
		for item in self.items:
			textFont(fonts['normal'])
			textSize(8)
			fill(255)
			text(item['ch'], 10, yoff)
			fill(113, 169, 231)
			text(item['title'], 26, yoff)
			textFont(fonts['tiny'])
			yoff += 1
			for line in textwrap.wrap(item['desc'], 24):
				yoff += 7
				text(line, 26, yoff)
			yoff += 10

		yoff += 12
	
		textAlign(CENTER)	
		text("Press <Q> to quit to menu", lwidth/2, yoff)
		yoff += 7
		text("or any other key to resume", lwidth/2, yoff)
		textFont(fonts['normal'])

	def keyPressed(self):
		if self.ended: return
		# wait a second before registering keypresses
		if millis() > self.init_millis + 250:
			if key == ord('q'):
				self.ended = True
				self.manager.add_state(BlockTransitionState(self, TitleScreenState()))
			else:
				self.manager.remove_state(self)
				self.callback()

	def state_removed(self):
		if self.ended:
			self.manager.remove_state(self.to_remove)

def ordinal(num):
	nums = str(num)
	if len(nums) >= 2 and int(nums[-2:]) >= 11 and int(nums[-2:]) <= 19:
		suff = "th"
	elif nums[-1] == "1": suff = "st"
	elif nums[-1] == "2": suff = "nd"
	elif nums[-1] == "3": suff = "rd"
	else: suff = "th"
	return nums + suff

def update_score_manager(mode, words_formed, letters_used, score, depth,
		seconds_elapsed, callback):
	score_manager.update_stats(mode, words_formed=len(words_formed),
			letters_used=letters_used, depth=depth,
			seconds_elapsed=seconds_elapsed)
	score_manager.record_game(mode, score, depth)
	score_manager.record_words(mode, words_formed)
	callback()

def text_left_of_center(s, yoff):
	w = textWidth(s)
	text(s, lwidth / 2 - w - 2, yoff)

class GameOverState(GameState):
	def __init__(self, to_remove, mode="none", words_formed=[],
			letters_used=0, elapsed_seconds=0, score=0, depth=0):
		self.init_millis = millis()
		self.to_remove = to_remove
		self.ended = False
		self.slice_size = 12
		self.yoff = lheight + 10
		self.xoff = (lwidth / 2) - (images['gameover'].width / 2)
		self.mode = mode
		self.minsec = minutes_seconds(elapsed_seconds)
		self.words_formed = words_formed
		self.letters_used = letters_used
		self.score = score
		self.depth = depth
		self.is_manager_updated = False
		self.is_remote_game_recorded = False
		self.is_remote_words_recorded = False
		self.is_tween_done = False
		self.will_update_remote = False
		self.got_remote_scores = False
		self.uniques = []
		self.substate_display_idx = 0
		self.leaderboard = {}
		self.high_score = 0
		self.best_depth = 0.0
		self.keypress_count = 0
		self.best_word = {'word': 'n/a', 'score': 'n/a'}
		self.mode_titles = {
			'challenge_easy': 'Arcade (Easy)',
			'challenge_norm': 'Arcade (Norm)',
			'challenge_hard': 'Arcade (Hard)',
			'timetrial_90': 'Timed (90sec)',
			'timetrial_240': 'Timed (4min)'
		}
		T.addTween(self, yoff=lheight * -1, tweenTime=1000,
				tweenType=T.IN_OUT_QUAD, onCompleteFunction=self.tween_done)
		self.words_formed.sort(key=lambda x: x['score'])
		if score_manager.get_setting('use_remote') is True:
			self.will_update_remote = True
		self.my_user_id = score_manager.get_user_id()
		thread = Thread(target=update_score_manager, args=(mode, words_formed,
				letters_used, score, depth, elapsed_seconds, self.manager_updated,))
		thread.start()

	def tween_done(self):
		self.is_tween_done = True
		self.check_ready_to_display_scores()

	def manager_updated(self):
		self.get_high_scores()
		self.is_manager_updated = True
		self.check_ready_to_display_scores()

	def check_ready_to_display_scores(self):
		if self.is_tween_done is True and self.is_manager_updated is True:
			self.high_score_display = HighScoreDisplayState(self)
			self.manager.add_state(self.high_score_display)
			if self.will_update_remote:
				self.unique_words_display = UniqueWordsDisplayState()
				self.manager.add_state(self.unique_words_display)
				remote_manager.record_game(self.my_user_id, self.mode, self.score,
						self.depth, self.remote_game_recorded)

	def get_high_scores(self):
		high_scores = score_manager.get_high_scores(mode=self.mode, count=1)
		if len(high_scores) > 0:
			self.high_score = high_scores[0].score
		else:
			self.high_score = 0
		self.high_score = max(self.score, self.high_score)
		best_depths = score_manager.get_best_depths(mode=self.mode, count=1)
		if len(best_depths) > 0:
			self.best_depth = best_depths[0].depth
		else:
			self.best_depth = 0
		self.best_depth = max(self.depth, self.best_depth)
		best_words = score_manager.get_best_words(mode=self.mode, count=1)
		if len(best_words) > 0:
			if best_words[0].score > self.words_formed[0]['score']:
				self.best_word = {'word': best_words[0].word,
						'score': best_words[0].score}
			else:
				self.best_word = self.words_formed[0]
		else:
			self.best_word = None

	def remote_high_scores_received(self, high_score, best_depth, best_word,
			best_word_score):
		self.global_high_score = high_score
		self.global_best_depth = best_depth
		self.global_best_word = best_word
		self.global_best_word_score = best_word_score
		self.got_remote_scores = True

	def remote_game_recorded(self, success, leaderboard):
		self.is_remote_game_recorded = True
		self.leaderboard = leaderboard
		remote_manager.record_words(self.my_user_id, self.mode, self.words_formed,
				self.remote_words_recorded)

	def remote_words_recorded(self, success, uniques):
		self.is_remote_words_recorded = True
		self.uniques = uniques
		if getattr(self, 'unique_words_display', None):
			self.unique_words_display.set_words(uniques)
		remote_manager.get_high_scores(self.mode, self.remote_high_scores_received)

	def draw(self):
		imageMode(CORNER)

		fill(0, 192)
		noStroke()
		rect(0, 0, lwidth, lheight)
		fill(0, 32)
		for i in range(0, lheight, 2):
			rect(0, float(i), lwidth, 1)

		tint(0, 128)
		image(images['gameover_shadow'], self.xoff + 3, self.yoff + 3)
		noTint()
		image(images['gameover'], self.xoff, self.yoff)
		xpos = (frameCount % images['gameoverhl'].width) * 4
		sdiff = images['gameoverhl'].width - xpos
		slice = min(self.slice_size, sdiff)
		copy(images['gameoverhl'],
				xpos, 0, slice, images['gameoverhl'].height, 
				(self.xoff + xpos) * game_scale, self.yoff * game_scale,
				slice * game_scale, images['gameoverhl'].height * game_scale)
		textFont(fonts['normal'])

	def keyPressed(self):
		if self.ended: return
		if key == CODED: return
		if millis() > self.init_millis + 500:
			self.keypress_count += 1
			if self.keypress_count == 1 and getattr(self, 'unique_words_display',
					None):
				T.addTween(self.high_score_display, xoff=-lwidth, tweenTime=1000)
				T.addTween(self.unique_words_display, xoff=-lwidth, tweenTime=1000)
			else:
				self.ended = True
				nextstate = TitleScreenState()
				self.manager.add_state(BlockTransitionState(self, nextstate))

	def state_removed(self):
		if getattr(self, 'unique_words_display', None):
			self.manager.remove_state(self.unique_words_display)
		if getattr(self, 'high_score_display', None):
			self.manager.remove_state(self.high_score_display)
		self.manager.remove_state(self.to_remove)

class HighScoreDisplayState(GameState):
	def __init__(self, game_over):
		self.game_over = game_over
		self.xoff = 0

	def draw(self):
		pushMatrix()
		translate(self.xoff, 0)

		fill(255)
		textAlign(LEFT)
		textFont(fonts['tiny'])
		yoff = lheight * 0.375

		fill(104, 234, 161)
		text_left_of_center("Score", yoff)
		fill(255)
		text("%d" % self.game_over.score, lwidth/2 + 2, yoff)

		yoff += 7
		fill(104, 234, 161)
		text_left_of_center("Your best", yoff)
		fill(255)
		text("%d" % self.game_over.high_score, lwidth/2 + 2, yoff)

		if self.game_over.will_update_remote:
			yoff += 7
			global_high_score = ("%d" % self.game_over.global_high_score) \
				if self.game_over.got_remote_scores \
				else ("|/-\\"[(int(frameCount * 0.3) % 4)])
			global_score_percentile = (
				ordinal(self.game_over.leaderboard['score']['percentile'])) \
				if self.game_over.got_remote_scores \
				else ("|/-\\"[(int(frameCount * 0.3) % 4)])
			fill(104, 234, 161)
			text_left_of_center("Global best", yoff)
			fill(255)
			text(global_high_score, lwidth/2 + 2, yoff)
			yoff += 7
			fill(104, 234, 161)
			text_left_of_center("Percentile", yoff)
			fill(255)
			text(global_score_percentile, lwidth/2 + 2, yoff)

		yoff += 12
		fill(104, 234, 161)
		text_left_of_center("Depth", yoff)
		fill(255)
		text("%.2f ab" % self.game_over.depth, lwidth/2 + 2, yoff)

		yoff += 7
		fill(104, 234, 161)
		text_left_of_center("Your best", yoff)
		fill(255)
		text("%.2f ab" % self.game_over.best_depth, lwidth/2 + 2, yoff)
		if self.game_over.will_update_remote:
			global_best_depth = ("%.2f ab" % self.game_over.global_best_depth) \
				if self.game_over.got_remote_scores \
				else ("|/-\\"[(int(frameCount * 0.3) % 4)])
			global_depth_percentile = (
				ordinal(self.game_over.leaderboard['depth']['percentile'])) \
				if self.game_over.got_remote_scores \
				else ("|/-\\"[(int(frameCount * 0.3) % 4)])
			yoff += 7
			fill(104, 234, 161)
			text_left_of_center("Global best", yoff)
			fill(255)
			text(global_best_depth, lwidth/2 + 2, yoff)
			yoff += 7
			fill(104, 234, 161)
			text_left_of_center("Percentile", yoff)
			fill(255)
			text(global_depth_percentile, lwidth/2 + 2, yoff)

		yoff += 12
		fill(104, 234, 161)
		text_left_of_center("Words formed", yoff)
		fill(255)
		text("%d" % len(self.game_over.words_formed), lwidth/2 + 2, yoff)

		yoff += 7
		fill(104, 234, 161)
		text_left_of_center("Letters used", yoff)
		fill(255)
		text("%d" % self.game_over.letters_used, lwidth/2 + 2, yoff)

		yoff += 7
		fill(104, 234, 161)
		text_left_of_center("Time", yoff)
		fill(255)
		text("%d:%02d" % (self.game_over.minsec[0], self.game_over.minsec[1]),
				lwidth/2 + 2, yoff)

		yoff += 7
		fill(104, 234, 161)
		text_left_of_center("Mode", yoff)
		fill(255)
		text(self.game_over.mode_titles[self.game_over.mode], lwidth/2 + 2, yoff)

		if len(self.game_over.words_formed) > 0:
			yoff += 12
			round_best_word = self.game_over.words_formed[-1]
			fill(104, 234, 161)
			text_left_of_center("Best word", yoff)
			fill(255)
			text(round_best_word['word'], lwidth/2 + 2, yoff)
			yoff += 7
			text("%d pts" % round_best_word['score'], lwidth/2 + 2, yoff)
			yoff += 7
			fill(104, 234, 161)
			text_left_of_center("All-time best", yoff)
			fill(255)
			text(self.game_over.best_word['word'], lwidth/2 + 2, yoff)
			yoff += 7
			text("%d pts" % self.game_over.best_word['score'], lwidth/2 + 2, yoff)
			if self.game_over.will_update_remote:
				global_best_word_score = ("%d pts" % self.game_over.global_best_word_score) \
					if self.game_over.got_remote_scores \
					else ("|/-\\"[(int(frameCount * 0.3) % 4)])
				global_best_word = (self.game_over.global_best_word) \
					if self.game_over.got_remote_scores \
					else ("|/-\\"[(int(frameCount * 0.3) % 4)])
				yoff += 7
				fill(104, 234, 161)
				text_left_of_center("Global best", yoff)
				fill(255)
				text(global_best_word, lwidth/2 + 2, yoff)
				yoff += 7
				text("%s" % global_best_word_score, lwidth/2 + 2, yoff)

		if self.game_over.will_update_remote:
			yoff += 12
			if not(self.game_over.is_remote_game_recorded) and \
					not(self.game_over.is_remote_words_recorded):
				txt = "." * (int(frameCount * 0.3) % 4)
			else:
				txt = "done!"
			text("Sending scores: " + txt, 8, yoff)
		textFont(fonts['normal'])
		popMatrix()

class UniqueWordsDisplayState(GameState):
	def __init__(self):
		self.words = None
		self.cols = None
		self.xoff = lwidth
	def set_words(self, words):
		self.words = words
		part = len(words) * 0.3333
		self.cols = [words[:int(part)+1], words[int(part)+1:int(part*2)+1],
				words[int(part*2)+1:]]
	def draw(self):
		pushMatrix()
		translate(self.xoff, 0)

		fill(255)
		textFont(fonts['tiny'])
		yoff = lheight * 0.4

		fill(104, 234, 161)
		textAlign(CENTER)
		text("You're the first player", lwidth/2, yoff)
		yoff += 7
		text("in the world to dig up", lwidth/2, yoff)
		yoff += 7
		text("these words:", lwidth/2, yoff)

		if self.words is None:
			yoff += 12
			fill(255)
			textAlign(CENTER)
			text("waiting...", lwidth/2, yoff)
		else:
			textAlign(LEFT)
			fill(255)
			yoff += 12
			if len(self.words) == 0:
				textAlign(CENTER)
				text("NONE! Better luck next time!", lwidth/2, yoff)
			else:
				for i, col in enumerate(self.cols):
					for j, word in enumerate(col):
						text(word, 10 + (i * 51), yoff+(j * 7))

		textFont(fonts['normal'])
		popMatrix()

class BlockTransitionState(GameState):
	def __init__(self, currentstate, nextstate):
		self.nextstate = nextstate
		self.currentstate = currentstate
		self.init_millis = millis()
		self.xsize = 8
		self.ysize = 12
		self.rectxsize = lwidth / self.xsize
		self.rectysize = lheight / self.ysize
		self.mode = 'in'
		
	def draw(self):
		imageMode(CORNER)
		if self.mode == 'in':
			to_draw = (millis() - self.init_millis) / 10
			drawn = 0
			for y in range(self.ysize):
				for x in range(self.xsize):
					pushMatrix()
					translate(x * self.rectxsize, y * self.rectysize)
					noStroke()
					fill(0, 64)
					rect(0, 0, self.rectxsize, self.rectysize)
					image(images['cyanblock'], 0, 0)
					popMatrix()
					drawn += 1
					if drawn >= to_draw: break
				if drawn >= to_draw: break
			if to_draw >= self.xsize * self.ysize:
				# remove the previous state, or the set of states with the given
				# tag
				if isinstance(self.currentstate, GameState):
					self.manager.remove_state(self.currentstate)
				elif isinstance(self.currentstate, basestring):
					self.manager.remove_tagged(self.currentstate)
				self.manager.add_state_beneath(self, self.nextstate)
				self.mode = 'out'
				self.init_millis = millis()
		elif self.mode == 'out':
			to_draw = (self.xsize * self.ysize) - ((millis() - self.init_millis) / 10)
			drawn = 0
			for x in range(self.xsize):
				for y in range(self.ysize):
					pushMatrix()
					translate(x * self.rectxsize, y * self.rectysize)
					noStroke()
					fill(0, 64)
					rect(0, 0, self.rectxsize, self.rectysize)
					image(images['cyanblock'], 0, 0)
					popMatrix()
					drawn += 1
					if drawn >= to_draw: break
				if drawn >= to_draw: break
			if to_draw <= 0:
				self.manager.remove_state(self)

class ImageSprite(object):
	def __init__(self, img, x, y, a=None, mode=CENTER):
		self.img = img
		self.x = x
		self.y = y
		self.a = a
		self.mode = mode 
		self.recalc_box()
	def draw(self):
		imageMode(self.mode)
		if self.a: tint(255, self.a)
		image(self.img, self.x, self.y)
		if self.a: noTint()
	def recalc_box(self):
		self.box_width = self.img.width
		self.box_height = self.img.height
		if self.mode is CENTER:
			self.box_x = self.x - (self.img.width / 2)
			self.box_y = self.y - (self.img.height / 2)
		else:
			self.box_x = self.x
			self.box_y = self.y
	def mouse_over(self):
		if lmouseX() > self.box_x - 2 \
				and lmouseX() < self.box_x + self.box_width + 2\
				and lmouseY() > self.box_y - 2\
				and lmouseY() < self.box_y + self.box_height + 2:
			return True
		return False

class TintCycleImageSprite(object):
	def __init__(self, img, x, y, a=None, mode=CENTER, delay=200, tints=None,
			shadow=False):
		self.img = img
		self.x = x
		self.y = y
		if a is None:
			self.a = 255
		else:
			self.a = a
		self.mode = mode
		self.delay = delay
		if tints is None:
			self.tints = [(255,255,255)]
		else:
			self.tints = tints
		self.last_change = millis()
		self.tints_idx = 0
		self.shadow = shadow
	def draw(self):
		imageMode(self.mode)
		if self.shadow:
			if self.a:
				tint(0, self.a * 0.5)
			else:
				tint(0, 128)
			image(self.img, self.x+2, self.y+2)
			noTint()
		if self.a:
			tint(*(tuple(self.tints[self.tints_idx]) + (self.a,)))
		else:
			tint(*self.tints[self.tints_idx])
		image(self.img, self.x, self.y)
		noTint()
		if millis() > self.last_change + self.delay:
			self.last_change = millis()
			self.tints_idx += 1
			if self.tints_idx > len(self.tints) - 1:
				self.tints_idx = 0

class ShadowStringSprite(object):
	def __init__(self, content, x, y, textsize=8, r=255, g=255, b=255, a=255,
			align=CENTER):
		self.content = content
		self.x = x
		self.y = y
		self.r = r
		self.g = g
		self.b = b
		self.a = a
		self.textsize = textsize
		self.align = align
		self.recalc_box()
	def set_font(self): pass
	def unset_font(self): pass
	def draw(self):
		self.set_font()
		textSize(self.textsize)
		textAlign(self.align)
		fill(0, 0, 0, self.a / 2.0)
		text(self.content, self.x + self.textsize * 0.2,
			self.y + self.textsize * 0.2)
		fill(self.r, self.g, self.b, self.a)
		text(self.content, self.x, self.y)
		self.unset_font()
	def recalc_box(self):
		self.set_font()
		textSize(self.textsize)
		self.box_width = textWidth(self.content)
		self.box_height = self.textsize
		if self.align == CENTER:
			self.box_x = self.x - (self.box_width / 2)
		else:
			self.box_x = self.x
		self.box_y = self.y - self.textsize
		self.unset_font()
	def mouse_over(self):
		if lmouseX() > self.box_x - 2 \
				and lmouseX() < self.box_x + self.box_width + 2\
				and lmouseY() > self.box_y - 2\
				and lmouseY() < self.box_y + self.box_height + 2:
			return True
		return False

class TinyShadowStringSprite(ShadowStringSprite):
	def set_font(self):
		textFont(fonts['tiny'])
	def unset_font(self):
		textFont(fonts['normal'])

class TitleScreenState(GameState):

	def __init__(self):
		self.logo_scale = 10.0

	def tween_finished(self):
		yoff = lheight * 0.65
		menustate = MainMenuState(yoff)
		self.manager.add_state_above(self, menustate)
		sound_manager.set_music('awake.ogg')

	def state_added(self):
		self.add_tag("titlescreen_group")
		T.addTween(self, logo_scale=-9.0, tweenTime=1000, tweenType=T.IN_OUT_QUAD,
			onCompleteFunction=self.tween_finished)

	def draw(self):
		imageMode(CORNER)
		bgheight = images['background'].height
		bgpos = (frameCount * 0.125 % bgheight) * -1
		image(images['background'], 0, bgpos)
		if bgheight + bgpos < lheight:
			image(images['background'], 0, (bgpos+bgheight))
		imageMode(CENTER)
		tint(0, 128)
		image(images['logo_shadow'], lwidth * 0.5 - 3, lheight * 0.3 - 3)
		noTint()
		pushMatrix()
		translate(lwidth * 0.5, lheight * 0.3)
		scale(self.logo_scale, self.logo_scale)
		image(images['logo'], 0, 0)
		popMatrix()

class MainMenuState(MenuState):
	def state_added(self):
		self.add_tag("titlescreen_group")
		yoff = self.yoff
		def tutorial_f():
			nextstate = TutorialModePlayfieldState(8, 12, checker, ngram_model)
			self.manager.add_state(BlockTransitionState("titlescreen_group",
					nextstate))
		self.add_item(
				{'spr': ShadowStringSprite("Tutorial", lwidth/2, yoff,
						align=CENTER), 'action': tutorial_f, 'text': 'Play this first'})
		def arcade_f():
			self.manager.mute(self)
			nextstate = ArcadeMenuState(yoff, last=self)
			self.manager.add_state(nextstate)
		self.add_item(
				{'spr': ShadowStringSprite("Arcade", lwidth/2, yoff + 12,
						align=CENTER), 'action': arcade_f, 'text': 'Dig to survive'})
		def time_f():
			self.manager.mute(self)
			nextstate = TimeTrialMenuState(yoff, last=self)
			self.manager.add_state(nextstate)
		self.add_item(
				{'spr': ShadowStringSprite("Time Trial", lwidth/2, yoff + 24,
						align=CENTER), 'action': time_f,
						'text': 'The clock is ticking'})
		def quest_f():
			nextstate = QuestModePlayfieldState(8, 12, checker, ngram_model)
			self.manager.add_state(BlockTransitionState("titlescreen_group",
					nextstate))
		tup = quest_manager.completion_tuple()
		self.add_item(
				{'spr': ShadowStringSprite("Quest Mode", lwidth/2, yoff + 36,
						align=CENTER), 'action': quest_f,
						'text': 'Completed %d/%d challenges' % tup})
		def more_menu_f():
			self.manager.mute(self)
			self.manager.add_state(MoreMenuState(yoff, last=self))
		self.add_item({'spr': ShadowStringSprite("More...", lwidth/2, yoff + 52,
				align=CENTER), 'action': more_menu_f,
				'text': 'Settings, credits, scores'})

class MoreMenuState(MenuState):
	def state_added(self):
		self.add_tag("titlescreen_group")
		yoff = self.yoff
		def settings_scores_f():
			self.manager.mute(self)
			self.manager.add_state(SettingsMenuState(yoff, last=self))
		self.add_item(
				{'spr': ShadowStringSprite("Settings", lwidth/2, yoff,
						align=CENTER),
						'action': settings_scores_f,
						'text': 'Adjust your preferences'})
		def records_f():
			nextstate = RecordsState()
			self.manager.add_state(BlockTransitionState("titlescreen_group",
					nextstate))
		self.add_item(
				{'spr': ShadowStringSprite("Records", lwidth/2, yoff + 12,
						align=CENTER),
						'action': records_f,
						'text': 'Review your scores'})
		def cuts_f():
			nextstate = ChoiceCutsState()
			self.manager.add_state(BlockTransitionState("titlescreen_group",
					nextstate))
		self.add_item(
				{'spr': ShadowStringSprite("Lexcellence", lwidth/2, yoff + 24,
						align=CENTER),
						'action': cuts_f,
						'text': 'Your best words'})
		def credits_f():
			nextstate = CreditsMenuState()
			self.manager.add_state(BlockTransitionState("titlescreen_group",
					nextstate))
		self.add_item(
				{'spr': ShadowStringSprite("Credits", lwidth/2, yoff + 36,
						align=CENTER),
						'action': credits_f,
						'text': 'Who made this?'})
		self.add_item(
				{'spr': ShadowStringSprite("Back", lwidth/2, yoff+52, align=CENTER),
					'action': self.back, 'text': "Return to menu"})

class CreditsMenuState(MenuState):
	def state_added(self):
		def menu_f():
			nextstate = TitleScreenState()
			self.manager.add_state(BlockTransitionState(self, nextstate))
		self.add_item({'spr': ShadowStringSprite("Menu", lwidth / 2, lheight - 8,
				align=CENTER), 'action': menu_f})
	def draw(self):
		imageMode(CORNER)
		image(images['background'], 0, 0)
		fill(0, 64)
		noStroke()
		rect(0, 0, lwidth, lheight)
		fill(0, 32)
		for i in range(0, lheight, 2):
			rect(0, float(i), lwidth, 1)

		yoff = 75
		fill(255)
		imageMode(CENTER)
		tint(0, 64)
		image(images['credits_shadow'], lwidth/2 + 3, lheight * 0.15 + 3)
		noTint()
		image(images['credits'], lwidth/2, lheight * 0.15)
		textFont(fonts['tiny'])
		textAlign(CENTER)
		text("Design, programming,", lwidth/2, yoff)
		yoff += 7
		text("graphics and music", lwidth/2, yoff)
		yoff += 7
 		text("by Allison Parrish", lwidth/2, yoff)
		yoff += 7
		fill(104, 234, 161)
		text("http://lexcavator.com", lwidth/2, yoff)
		yoff += 14
		textAlign(LEFT)
		fill(255)
		text("Includes components from:", 5, yoff)
		yoff += 12
		text("processing.py", 5, yoff)
		yoff += 7
		fill(104, 234, 161)
		text("github.com/jdf/processing.py", 5, yoff)
		yoff += 12
		fill(255)
		text("SoundSystem", 5, yoff)
		yoff += 7
		fill(104, 234, 161)
		text("paulscode.com", 5, yoff)
		yoff += 12
		fill(255)
		text("codeman38 fonts", 5, yoff)
		yoff += 7
		fill(104, 234, 161)
		text("www.zone38.net/font", 5, yoff)
		yoff += 14
		fill(255)
		text("See credits.txt for complete", 5, yoff)
		yoff += 7
		text("credits and license info.", 5, yoff)
		yoff += 14
		fill(104, 234, 161)
		textAlign(CENTER)
		text("Special thanks to the Warp", lwidth/2, yoff)
		yoff += 7
		text("Skip Google Wave crew.", lwidth/2, yoff)
		textFont(fonts['normal'])
		super(CreditsMenuState, self).draw()

class ArcadeMenuState(MenuState):
	def state_added(self):
		self.add_tag("titlescreen_group")
		yoff = self.yoff
		def add_with_level(lev):
			nextstate = ChallengeModePlayfieldState(8, 12, checker,
				ngram_model, lev)
			self.manager.add_state(BlockTransitionState("titlescreen_group",
					nextstate))
		self.add_item(
				{'spr': ShadowStringSprite("Cakewalk", lwidth/2, yoff, align=CENTER),
					'action': lambda: add_with_level('easy'),
					'text': 'Easy mode'})
		self.add_item(
				{'spr': ShadowStringSprite("Reasonable", lwidth/2, yoff+12,
						align=CENTER),
					'action': lambda: add_with_level('norm'),
					'text': 'Normal'})
		self.add_item(
				{'spr': ShadowStringSprite("Unyielding", lwidth/2, yoff+24,
						align=CENTER),
					'action': lambda: add_with_level('hard'),
					'text': 'Hard'})
		self.add_item(
				{'spr': ShadowStringSprite("Back", lwidth/2, yoff+42, align=CENTER),
					'action': self.back, 'text': "Return to menu"})

class TimeTrialMenuState(MenuState):
	def state_added(self):
		self.add_tag("titlescreen_group")
		yoff = self.yoff
		def add_with_time(secs):
			nextstate = TimeTrialPlayfieldState(8, 12, checker, ngram_model, secs)
			self.manager.add_state(BlockTransitionState("titlescreen_group",
					nextstate))
		self.add_item(
				{'spr': ShadowStringSprite("90 Seconds", lwidth/2, yoff, align=CENTER),
					'action': lambda: add_with_time(90), 'text': 'Quickie'})
		self.add_item(
				{'spr': ShadowStringSprite("4 Minutes", lwidth/2, yoff+12,
					align=CENTER), 'action': lambda: add_with_time(240),
					'text': 'Long haul'})
		self.add_item(
				{'spr': ShadowStringSprite("Back", lwidth/2, yoff+32, align=CENTER),
					'action': self.back, 'text': "Return to menu"})

class SettingsMenuState(MenuState):
	def state_added(self):
		self.add_tag("titlescreen_group")
		yoff = self.yoff

		def toggle_music(spr):
			if score_manager.get_setting('music') in (None, True):
				score_manager.set_setting('music', False)
				sound_manager.set_play_music(False)
				spr.img = images['checkbox']
			else:
				score_manager.set_setting('music', True)
				sound_manager.set_play_music(True)
				spr.img = images['checked']
			score_manager.save()

		if score_manager.get_setting('music') in (None, True):
			music_check_spr = ImageSprite(images['checked'], 30, yoff-3)
		else:
			music_check_spr = ImageSprite(images['checkbox'], 30, yoff-3)
		self.add_item({'spr': music_check_spr,
				'action': lambda: toggle_music(music_check_spr),
				'text': 'Toggle music'})
		music_spr = ShadowStringSprite("Music", 40, yoff, align=LEFT)
		self.add_item({'spr': music_spr,
				'action': lambda: toggle_music(music_check_spr),
				'text': 'Toggle music'})

		def toggle_sfx(spr):
			if score_manager.get_setting('sfx') in (None, True):
				score_manager.set_setting('sfx', False)
				sound_manager.set_play_sfx(False)
				spr.img = images['checkbox']
			else:
				score_manager.set_setting('sfx', True)
				sound_manager.set_play_sfx(True)
				spr.img = images['checked']
			score_manager.save()

		if score_manager.get_setting('sfx') in (None, True):
			sfx_check_spr = ImageSprite(images['checked'], 30, yoff+9)
		else:
			sfx_check_spr = ImageSprite(images['checkbox'], 30, yoff+9)
		self.add_item({'spr': sfx_check_spr,
				'action': lambda: toggle_music(sfx_check_spr),
				'text': 'Toggle sound effects'})
		sfx_spr = ShadowStringSprite("Sound effects", 40, yoff+12, align=LEFT)
		self.add_item({'spr': sfx_spr, 'action': lambda: toggle_sfx(sfx_check_spr),
				'text': 'Toggle sound effects'})

		def toggle_remote(spr):
			if score_manager.get_setting('use_remote') in (None, True):
				score_manager.set_setting('use_remote', False)
				spr.img = images['checkbox']
			else:
				score_manager.set_setting('use_remote', True)
				spr.img = images['checked']
			self.update_checkmarks()
			score_manager.save()

		if score_manager.get_setting('use_remote') in (None, True):
			remote_check_spr = ImageSprite(images['checked'], 30, yoff+21)
		else:
			remote_check_spr = ImageSprite(images['checkbox'], 30, yoff+21)
		self.add_item({'spr': remote_check_spr,
				'action': lambda: toggle_remote(remote_check_spr),
				'text': 'Join in on global scoreboard'})
		remote_spr = ShadowStringSprite("Internet", 40, yoff+24, align=LEFT)
		self.add_item({'spr': remote_spr,
				'action': lambda: toggle_remote(remote_check_spr),
				'text': 'Join in on global scoreboard'})

		self.add_item(
				{'spr': ShadowStringSprite("Back", lwidth/2, yoff+44, align=CENTER),
					'action': self.back, 'text': "Return to menu"})

	def update_checkmarks(self):
		self.music_check = score_manager.get_setting('music')
		self.sound_check = score_manager.get_setting('sfx')
		self.remote_check = score_manager.get_setting('use_remote')

class ChoiceCutsState(MenuState):
	def state_added(self):
		word_recs = score_manager.get_best_words(None, count=42)
		words = [rec.word for rec in word_recs]
		part = len(words) * 0.3333
		self.cols = [words[:int(part)+1], words[int(part)+1:int(part*2)+1],
				words[int(part*2)+1:]]
		self.words = words
		def menu_f():
			nextstate = TitleScreenState()
			self.manager.add_state(BlockTransitionState(self, nextstate))
		self.add_item({'spr': ShadowStringSprite("Menu", lwidth / 2, lheight - 8,
				align=CENTER), 'action': menu_f})

	def draw(self):
		imageMode(CORNER)
		image(images['background'], 0, 0)
		fill(0, 64)
		noStroke()
		rect(0, 0, lwidth, lheight)
		fill(0, 32)
		for i in range(0, lheight, 2):
			rect(0, float(i), lwidth, 1)

		imageMode(CENTER)
		tint(0, 64)
		image(images['lexcellence_shadow'], lwidth/2 + 3, lheight * 0.21 + 3)
		noTint()
		image(images['lexcellence'], lwidth/2, lheight * 0.21)

		fill(255)
		textFont(fonts['tiny'])
		yoff = 104
		fill(104, 234, 161)
		textAlign(CENTER)
		text("Your best words", lwidth/2, yoff)
		yoff += 12

		textAlign(LEFT)
		fill(255)
		if len(self.words) == 0:
			textAlign(CENTER)
			text("No words found yet!", lwidth/2, yoff)
		else:
			for i, col in enumerate(self.cols):
				for j, word in enumerate(col):
					text(word, 8 + (i * 52), yoff+(j * 8))

		textFont(fonts['normal'])
		super(ChoiceCutsState, self).draw()

class RecordsState(MenuState):
	def state_added(self):
		self.modes = [
			{'title': 'All modes', 'search': None},
			{'title': 'Arcade (Easy)', 'search': 'challenge_easy'},
			{'title': 'Arcade (Normal)', 'search': 'challenge_norm'},
			{'title': 'Arcade (Hard)', 'search': 'challenge_hard'},
			{'title': 'Time Trial (90 sec)', 'search': 'timetrial_90'},
			{'title': 'Time Trial (4 min)', 'search': 'timetrial_240'},
			{'title': 'Quest Mode', 'search': 'quest'}]
		self.current_mode = 0
		self.switch_to(self.current_mode)
		def next_f():
			self.current_mode += 1
			if self.current_mode > len(self.modes)-1:
				self.current_mode = 0
			self.switch_to(self.current_mode)
		def menu_f():
			nextstate = TitleScreenState()
			self.manager.add_state(BlockTransitionState(self, nextstate))
		self.add_item({'spr': ShadowStringSprite("Next >", lwidth - (lwidth/3),
				lheight - 8, align=CENTER), 'action': next_f})
		self.add_item({'spr': ShadowStringSprite("Menu", lwidth / 3, lheight - 8,
				align=CENTER), 'action': menu_f})

	def switch_to(self, mode):
		modestr = self.modes[mode]['search']
		self.stats = score_manager.get_stats(modestr)
		self.hiscores = score_manager.get_high_scores(modestr, count=4)
		self.bestdepths = score_manager.get_best_depths(modestr, count=4)
		self.bestwords = score_manager.get_best_words(modestr, count=4)
		m, s = divmod(self.stats['seconds_elapsed'], 60)
		h, m = divmod(m, 60)
		self.total_time = ':'.join(["%02d" % x for x in (h, m, s)])

	def draw(self):
		imageMode(CORNER)
		image(images['background'], 0, 0)
		fill(0, 64)
		noStroke()
		rect(0, 0, lwidth, lheight)
		fill(0, 32)
		for i in range(0, lheight, 2):
			rect(0, float(i), lwidth, 1)

		yoff = 75
		fill(255)
		imageMode(CENTER)
		tint(0, 64)
		image(images['records_shadow'], lwidth/2 + 3, lheight * 0.1 + 3)
		noTint()
		image(images['records'], lwidth/2, lheight * 0.1)
		textFont(fonts['tiny'])
	
		textSize(8)
		textAlign(CENTER)
		fill(104, 234, 161)
		text(self.modes[self.current_mode]['title'], lwidth/2, 53)
		yoff = 65
		textFont(fonts['tiny'])
		textAlign(LEFT)

		if self.modes[self.current_mode]['search'] != 'quest':
			fill(104, 234, 161)
			text("Best scores", 4, yoff)
			fill(255)
			yoff += 7
			for i, score_rec in enumerate(self.hiscores):
				text("%d" % (i+1), 4, yoff)
				text("%07d" % score_rec.score, 12, yoff)
				text("%s" % score_rec.created_at.strftime("%m/%d/%y"), 106, yoff)
				yoff += 7

			yoff += 5
			fill(104, 234, 161)
			text("Best depths", 4, yoff)
			fill(255)
			yoff += 7
			for i, depth_rec in enumerate(self.bestdepths):
				text("%d" % (i+1), 4, yoff)
				text("%02.02f" % depth_rec.depth, 12, yoff)
				text("%s" % depth_rec.created_at.strftime("%m/%d/%y"), 106, yoff)
				yoff += 7

			yoff += 5
			fill(104, 234, 161)
			text("Best words", 4, yoff)
			fill(255)
			yoff += 7
			for i, word_rec in enumerate(self.bestwords):
				text("%d" % (i+1), 4, yoff)
				text("%s" % word_rec.word, 12, yoff)
				text("%06d" % word_rec.score, 64, yoff)
				text(word_rec.created_at.strftime("%m/%d/%y"), 106, yoff)
				yoff += 7

		yoff += 5
		fill(104, 234, 161)
		text_left_of_center("Words formed", yoff)
		fill(255)
		text("%d" % self.stats['words_formed'], lwidth/2+2, yoff)
		if self.stats['words_formed'] >= 1000:
			image(images['goldstar'],
				lwidth/2+5+textWidth("%d" % self.stats['words_formed']), yoff-3)
		yoff += 7
		fill(104, 234, 161)
		text_left_of_center("Letters used", yoff)
		fill(255)
		text("%d" % self.stats['letters_used'], lwidth/2+2, yoff)
		if self.stats['letters_used'] >= 5000:
			image(images['goldstar'],
				lwidth/2+5+textWidth("%d" % self.stats['letters_used']), yoff-3)
		yoff += 7
		fill(104, 234, 161)
		text_left_of_center("Total depth", yoff)
		fill(255)
		text("%2.02f ab" % self.stats['total_depth'], lwidth/2+2, yoff)
		if self.stats['total_depth'] >= 100.0:
			image(images['goldstar'],
				lwidth/2+5+textWidth("%2.02f ab" % self.stats['total_depth']), yoff-3)
		yoff += 7
		fill(104, 234, 161)
		text_left_of_center("Time played", yoff)
		fill(255)
		text("%s" % self.total_time, lwidth/2 + 2, yoff)
		if self.stats['seconds_elapsed'] >= 18000:
			image(images['goldstar'],
				lwidth/2+5+textWidth("%s" % self.total_time), yoff-3)
		yoff += 7
		fill(104, 234, 161)
		text_left_of_center("Games played", yoff)
		fill(255)
		text("%d" % (self.stats['games_played']), lwidth/2 + 2, yoff)
		if self.stats['games_played'] >= 100:
			image(images['goldstar'],
				lwidth/2+5+textWidth("%d" % self.stats['games_played']), yoff-3)

		textFont(fonts['normal'])
		super(RecordsState, self).draw()

def version_callback(remote_version):
	if remote_version > client_version:
		score_manager.set_setting('show_version_warning', True)

class VersionCheckState(MenuState):
	def __init__(self):
		self.lines = []
		super(VersionCheckState, self).__init__(self)

	def state_added(self):
		if score_manager.get_setting('use_remote') in (None, True):
			raw_osname = System.getProperty("os.name").lower()
			if 'mac' in raw_osname:
				osname = 'mac'
			elif 'win' in raw_osname:
				osname = 'win'
			elif 'nux' in raw_osname:
				osname = 'linux'
			remote_manager.get_version(osname, client_version, version_callback)
		if score_manager.get_setting('show_version_warning') is False:
			self.manager.add_state(TitleScreenState())
			self.manager.remove_state(self)
		else:
			score_manager.set_setting('show_version_warning', False)
			self.lines = [
				"Looks like your version",
				"of Lexcavator is out of",
				"date! Certain online",
				"features might not work",
				"as expected. Visit",
				"http://www.lexcavator.com",
				"to download the latest",
				"version."]

			yoff = 149
			def download_f():
				Desktop.getDesktop().browse(URI("http://www.lexcavator.com/?r=up"))
			try:
				if Desktop.getDesktop().isSupported(Desktop.Action.BROWSE):
					self.add_item({'spr': ShadowStringSprite("Open browser", lwidth / 2,
							yoff), 'action': download_f})
					yoff += 12
			except java.lang.UnsupportedOperationException:
				pass

			def continue_f():
				self.manager.remove_state(self)
				self.manager.add_state(TitleScreenState())
			self.add_item({'spr': ShadowStringSprite("Continue", lwidth / 2,
					yoff), 'action': continue_f})

	def draw(self):
		imageMode(CORNER)
		image(images['background'], 0, 0)
		fill(0, 64)
		noStroke()
		rect(0, 0, lwidth, lheight)
		fill(0, 32)
		for i in range(0, lheight, 2):
			rect(0, float(i), lwidth, 1)

		textSize(8)
		fill(255)
		textFont(fonts['tiny'])
		textAlign(LEFT)
		yoff = 79
		for line in self.lines:
			text(line, 8, yoff)
			yoff += 7
		textFont(fonts['normal'])
		super(VersionCheckState, self).draw()

class InternetCheckState(MenuState):

	def __init__(self):
		self.lines = []
		super(InternetCheckState, self).__init__(self)

	def state_added(self):
		if score_manager.get_setting('first_time') is False:
			self.manager.add_state(VersionCheckState())
			self.manager.remove_state(self)
			return
		score_manager.set_setting('first_time', False)

		def toggle_remote(spr):
			if score_manager.get_setting('use_remote') in (None, True):
				score_manager.set_setting('use_remote', False)
				spr.img = images['checkbox']
			else:
				score_manager.set_setting('use_remote', True)
				spr.img = images['checked']
			self.update_checkmarks()
			score_manager.save()

		if score_manager.get_setting('use_remote') in (None, True):
			remote_check_spr = ImageSprite(images['checked'], 43, 168)
		else:
			remote_check_spr = ImageSprite(images['checkbox'], 43, 168)
		self.add_item({'spr': remote_check_spr,
				'action': lambda: toggle_remote(remote_check_spr),
				'text': ''})
		remote_spr = ShadowStringSprite("Internet", 53, 171, align=LEFT)
		self.add_item({'spr': remote_spr,
				'action': lambda: toggle_remote(remote_check_spr),
				'text': ''})

		def continue_f():
			self.manager.remove_state(self)
			self.manager.add_state(VersionCheckState())
		self.add_item({'spr': ShadowStringSprite("Continue", lwidth / 2,
				188), 'action': continue_f})

		self.lines = [
			"Welcome to Lexcavator!",
			"",
			"Want to know how your",
			"skills stack up? Enable",
			"Internet access to",
			"participate in Lexcavator's",
			"global leaderboard.",
			"", 
			"Your scores will be",
			"submitted and stored",
			"anonymously.",
			"", 
			"(You can turn off Internet",
			"access later in the Settings",
			"menu.)"]

	def update_checkmarks(self):
		self.remote_check = score_manager.get_setting('use_remote')

	def draw(self):
		imageMode(CORNER)
		image(images['background'], 0, 0)
		fill(0, 64)
		noStroke()
		rect(0, 0, lwidth, lheight)
		fill(0, 32)
		for i in range(0, lheight, 2):
			rect(0, float(i), lwidth, 1)

		textSize(8)
		fill(255)
		textFont(fonts['tiny'])
		textAlign(LEFT)
		yoff = 54
		for line in self.lines:
			text(line, 8, yoff)
			yoff += 7
		textFont(fonts['normal'])
		super(InternetCheckState, self).draw()

def loadchecker(callback):
	global checker, ngram_model, score_manager, sound_manager, remote_manager,\
		 quest_manager

	# load image and sound resources
	print "loading images"
	images['cyanblock'] = loadImage('cyanblock.png')
	images['orangeblock'] = loadImage('orangeblock.png')
	images['narrow'] = loadImage('narrow.png')
	images['narrow_mask'] = loadImage('narrow_mask.png')
	images['nwarrow'] = loadImage('nwarrow.png')
	images['nwarrow_mask'] = loadImage('nwarrow_mask.png')
	images['greenblock'] = loadImage('greenblock.png')
	images['selector1'] = loadImage('selector1.png')
	images['selector2'] = loadImage('selector2.png')
	images['background'] = loadImage('background.png')
	images['logo'] = loadImage('logo_dbl.png')
	images['logo_shadow'] = loadImage('logo_shadow.png')
	images['paused'] = loadImage('pause.png')
	images['paused_shadow'] = loadImage('pause_shadow.png')
	images['gameover'] = loadImage('gameover.png')
	images['gameover_shadow'] = loadImage('gameover_shadow.png')
	images['gameoverhl'] = loadImage('gameoverhl.png')
	images['levelup'] = loadImage('levelup.png')
	images['levelup_shadow'] = loadImage('levelup_shadow.png')
	images['quest_progress'] = loadImage('quest_progress.png')
	images['quest_progress_shadow'] = loadImage('quest_progress_shadow.png')
	images['checked'] = loadImage('checked.png')
	images['checkbox'] = loadImage('checkbox.png')
	images['credits'] = loadImage('credits.png')
	images['credits_shadow'] = loadImage('credits_shadow.png')
	images['records'] = loadImage('records.png')
	images['records_shadow'] = loadImage('records_shadow.png')
	images['lexcellence'] = loadImage('lexcellence.png')
	images['lexcellence_shadow'] = loadImage('lexcellence_shadow.png')
	images['bonus'] = loadImage('bonus.png')
	images['goldstar'] = loadImage('goldstar.png')
	images['cursor'] = loadImage('cursor.png')
	fonts['tiny'] = createFont('pixelette.ttf', 8)

	print "loading wordchecker and ngram model"

	# word checker and ngram models
	checker = WordChecker()
	ngram_model = list()
	for i in range(4,ngramlength+1):
		mod = NGramModel(n=i)
		ngram_model.append(mod)

	loader = java.lang.ClassLoader.getSystemClassLoader()	
	for line in FileUtil.wrap(loader.getResourceAsStream('wordlist.txt')):
		line = line.strip()
		checker.feed(line)
		if len(line) <= 12:
			for mod in ngram_model:
				mod.feed(line)
	for mod in ngram_model:
		mod.finish()

	print "creating sound manager"
	sound_manager = SoundManager(ss)

	print "loading data into score manager"
	# directories for data persistence
	data_dir = userdir.get_user_data_dir("Lexcavator")
	if not os.path.exists(data_dir):
		os.makedirs(data_dir)
	score_manager = ScoreManager(os.path.join(data_dir))

	quest_manager = quests.QuestManager(score_manager)
	if len(quest_manager.get_quests()) == 0:
		quest_manager.initialize()
	quest_manager.restore()

	remote_manager = AsyncRemoteScoreManager(base_url=remote_baseurl)

	if score_manager.get_setting('music') is not False:
		sound_manager.set_play_music(True)
	else:
		sound_manager.set_play_music(False)
	if score_manager.get_setting('sfx') is not False:
		sound_manager.set_play_sfx(True)
	else:
		sound_manager.set_play_sfx(False)

	callback()

class LoadResourcesState(GameState):
	def __init__(self, nextstate):
		self.nextstate = nextstate
		self.thread = Thread(target=loadchecker, args=(self.checker_loaded,))
		self.thread.start()
	def draw(self):
		if self.thread.isAlive():
			textSize(8)
			textAlign(LEFT)
			fill(227, 114, 20)
			progress = '.' * (int(frameCount * 0.08) % 4)
			text("loading%s" % progress, lwidth/2 - 40, lheight/2)
	def checker_loaded(self):
		if isinstance(self.nextstate, type):
			self.manager.add_state(self.nextstate())
		elif isinstance(self.nextstate, GameState):
			self.manager.add_state(self.nextstate)
		self.manager.remove_state(self)

class Sketch(GameStateManager):
	def setup(self):
		global game_scale
		screen = Toolkit.getDefaultToolkit().getScreenSize()
		game_scale = (screen.height-50) / lheight
		frameRate(30)
		size(lwidth * game_scale, lheight * game_scale)
		background(0)
		font = createFont("pcsenior.ttf", 16)
		fonts['normal'] = font
		textFont(font)
		frame.setTitle("Lexcavator")
		frame.setIconImage(loadImage('favicon.png').getImage())
		self.add_state(
			LoadResourcesState(nextstate=InternetCheckState))
		self.s = millis()
		self.cursor = loadImage('cursor.png')
		self.show_fps = False
		self.software_cursor = False
		self.recording = False
	def update(self, delta):
		super(Sketch, self).update(delta)
	def draw(self):
		background(0)
		tm = float(millis())
		delta = tm - self.s
		self.s = tm
		T.update(delta)
		self.update(delta)
		scale(game_scale, game_scale)
		super(Sketch, self).draw()
		if self.show_fps:
			textFont(fonts['tiny'])
			textAlign(RIGHT)
			text("%0.0f" % frameRate, lwidth-1, 6)
			textFont(fonts['normal'])
		if self.software_cursor:
			noCursor()
			imageMode(CORNER)
			image(self.cursor, mouseX/game_scale, mouseY/game_scale)
		if self.recording:
			saveFrame("lexcavator-####.png")
	def keyPressed(self):
		if key == ord('.'):
			self.show_fps = not(self.show_fps)
		elif key == ord('M'):
			self.software_cursor = not(self.software_cursor)
		elif key == ord('R'):
			self.recording = not(self.recording)
		super(Sketch, self).keyPressed()

T = Tweener()
sketch = Sketch()
images = dict()
sounds = dict()
fonts = dict()

SoundSystemConfig.addLibrary(LibraryLWJGLOpenAL)
SoundSystemConfig.setCodec("wav", CodecWav)
SoundSystemConfig.setCodec("ogg", CodecJOrbis)
ss = SoundSystem()

def setup():
	sketch.setup()
def draw():
	sketch.draw()
#	if ss and frameCount % 3 == 0:
#		ss.checkFadeVolumes()
def mousePressed():
	sketch.mousePressed()
def mouseReleased():
	sketch.mouseReleased()
def keyPressed():
	if key == 27:
		this.key = '\0'
	elif key == ord('p'):
		saveFrame("lexcavator-screenshot-####.png")
	sketch.keyPressed()
def keyReleased():
	sketch.keyReleased()
def stop():
	ss.cleanup()

class ShutdownTask(java.lang.Thread):
	def run(self):
		stop()
java.lang.Runtime.getRuntime().addShutdownHook(ShutdownTask())
